@@ -1,7 +1,7 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8; coding: utf-8 -*-
  *  gtksourcelanguagemanager.c
  *
- *  Copyright (C) 2003 - Paolo Maggi <paolo.maggi@polito.it>
+ *  Copyright (C) 2003-2007 - Paolo Maggi <paolo.maggi@polito.it>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Library General Public License as published by
@@ -34,15 +34,18 @@
 
 enum {
 	PROP_0,
-	PROP_SEARCH_PATH
+	PROP_SEARCH_PATH,
+	PROP_LANGUAGE_IDS
 };
 
 struct _GtkSourceLanguageManagerPrivate
 {
 	GHashTable	*language_ids;
-	GSList 		*available_languages;
-	char	       **lang_dirs;
-	char		*rng_file;
+
+	gchar	       **lang_dirs;
+	gchar		*rng_file;
+
+	gchar          **ids; /* Cache the IDs of the available languages */
 };
 
 G_DEFINE_TYPE (GtkSourceLanguageManager, gtk_source_language_manager, G_TYPE_OBJECT)
@@ -86,6 +89,10 @@ gtk_source_language_manager_get_property (GObject 	*object,
 			g_value_set_boxed (value, gtk_source_language_manager_get_search_path (lm));
 			break;
 
+		case PROP_LANGUAGE_IDS:
+			g_value_set_boxed (value, gtk_source_language_manager_get_language_ids (lm));
+			break;
+
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -102,8 +109,8 @@ gtk_source_language_manager_finalize (GObject *object)
 	if (lm->priv->language_ids)
 		g_hash_table_destroy (lm->priv->language_ids);
 
-	g_slist_foreach (lm->priv->available_languages, (GFunc) g_object_unref, NULL);
-	g_slist_free (lm->priv->available_languages);
+	g_strfreev (lm->priv->ids);
+
 	g_strfreev (lm->priv->lang_dirs);
 	g_free (lm->priv->rng_file);
 
@@ -130,6 +137,15 @@ gtk_source_language_manager_class_init (GtkSourceLanguageManagerClass *klass)
 							     G_TYPE_STRV,
 							     G_PARAM_READWRITE));
 
+	g_object_class_install_property (object_class,
+					 PROP_LANGUAGE_IDS,
+					 g_param_spec_boxed ("language-ids",
+						 	     _("Language ids"),
+							     _("List of the ids of the available "
+							       "languages"),
+							     G_TYPE_STRV,
+							     G_PARAM_READABLE));
+
 	g_type_class_add_private (object_class, sizeof(GtkSourceLanguageManagerPrivate));
 }
 
@@ -138,9 +154,8 @@ gtk_source_language_manager_init (GtkSourceLanguageManager *lm)
 {
 	lm->priv = G_TYPE_INSTANCE_GET_PRIVATE (lm, GTK_TYPE_SOURCE_LANGUAGE_MANAGER,
 						GtkSourceLanguageManagerPrivate);
-	lm->priv = g_new0 (GtkSourceLanguageManagerPrivate, 1);
 	lm->priv->language_ids = NULL;
-	lm->priv->available_languages = NULL;
+	lm->priv->ids = NULL;
 	lm->priv->lang_dirs = NULL;
 	lm->priv->rng_file = NULL;
 }
@@ -148,51 +163,100 @@ gtk_source_language_manager_init (GtkSourceLanguageManager *lm)
 /**
  * gtk_source_language_manager_new:
  *
- * Creates a new language manager.
+ * Creates a new language manager. If you do not need more than one language
+ * manager or a private language manager instance then use
+ * gtk_source_language_manager_get_default() instead.
  *
  * Returns: a #GtkSourceLanguageManager.
- **/
+ */
 GtkSourceLanguageManager *
 gtk_source_language_manager_new (void)
 {
 	return g_object_new (GTK_TYPE_SOURCE_LANGUAGE_MANAGER, NULL);
 }
 
+/**
+ * gtk_source_language_manager_get_default:
+ *
+ * Returns the default #GtkSourceLanguageManager instance.
+ *
+ * Returns: a #GtkSourceLanguageManager. Return value is owned
+ * by GtkSourceView library and must not be unref'ed.
+ */
+GtkSourceLanguageManager *
+gtk_source_language_manager_get_default (void)
+{
+	static GtkSourceLanguageManager *instance;
+
+	if (instance == NULL)
+	{
+		instance = gtk_source_language_manager_new ();
+		g_object_add_weak_pointer (G_OBJECT (instance),
+					   (gpointer) &instance);
+	}
+
+	return instance;
+}
+
+static void
+notify_search_path (GtkSourceLanguageManager *mgr)
+{
+	g_object_notify (G_OBJECT (mgr), "search-path");
+	g_object_notify (G_OBJECT (mgr), "language-ids");
+}
+
 /**
  * gtk_source_language_manager_set_search_path:
  * @lm: a #GtkSourceLanguageManager.
  * @dirs: a %NULL-terminated array of strings or %NULL.
  *
- * Sets a list of language files directories for the given language manager.
- * @dirs == %NULL resets directories list to default.
- **/
+ * Sets the list of directories where the @lm looks for
+ * language files.
+ * If @dirs is %NULL, the search path is reset to default.
+ *
+ * <note>
+ *   <para>
+ *     At the moment this function can be called only before the
+ *     language files are loaded for the first time. In practice
+ *     to set a custom search path for a #GtkSourceLanguageManager,
+ *     you have to call this function right after creating it.
+ *   </para>
+ * </note>
+ */
 void
 gtk_source_language_manager_set_search_path (GtkSourceLanguageManager *lm,
 					     gchar                   **dirs)
 {
-	char **tmp;
+	gchar **tmp;
 
 	g_return_if_fail (GTK_IS_SOURCE_LANGUAGE_MANAGER (lm));
-	/* FIXME: so what do we do? rescan, refresh? */
-	g_return_if_fail (lm->priv->available_languages == NULL);
+
+	/* Search path cannot be changed in the list of available languages
+	 * as been already computed */
+	g_return_if_fail (lm->priv->ids == NULL);
 
 	tmp = lm->priv->lang_dirs;
-	lm->priv->lang_dirs = g_strdupv (dirs);
+
+	if (dirs == NULL)
+		lm->priv->lang_dirs = _gtk_source_view_get_default_dirs (LANGUAGE_DIR, TRUE);
+	else
+		lm->priv->lang_dirs = g_strdupv (dirs);
+
 	g_strfreev (tmp);
 
-	g_object_notify (G_OBJECT (lm), "search-path");
+	notify_search_path (lm);
 }
 
 /**
  * gtk_source_language_manager_get_search_path:
  * @lm: a #GtkSourceLanguageManager.
  *
- * Gets a list of language files directories for the given language manager.
+ * Gets the list directories where @lm looks for language files.
  *
  * Returns: %NULL-terminated array containg a list of language files directories.
- * It is owned by @lm and must not be modified or freed.
- **/
-gchar **
+ * The array is owned by @lm and must not be modified.
+ */
+G_CONST_RETURN gchar* G_CONST_RETURN *
 gtk_source_language_manager_get_search_path (GtkSourceLanguageManager *lm)
 {
 	g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE_MANAGER (lm), NULL);
@@ -200,7 +264,7 @@ gtk_source_language_manager_get_search_path (GtkSourceLanguageManager *lm)
 	if (lm->priv->lang_dirs == NULL)
 		lm->priv->lang_dirs = _gtk_source_view_get_default_dirs (LANGUAGE_DIR, TRUE);
 
-	return lm->priv->lang_dirs;
+	return (const gchar * const *)lm->priv->lang_dirs;
 }
 
 /**
@@ -210,7 +274,7 @@ gtk_source_language_manager_get_search_path (GtkSourceLanguageManager *lm)
  * Returns location of the RNG schema file for lang files version 2.
  *
  * Returns: path to RNG file. It belongs to %lm and must not be freed or modified.
- **/
+ */
 const char *
 _gtk_source_language_manager_get_rng_file (GtkSourceLanguageManager *lm)
 {
@@ -218,7 +282,7 @@ _gtk_source_language_manager_get_rng_file (GtkSourceLanguageManager *lm)
 
 	if (lm->priv->rng_file == NULL)
 	{
-		gchar **dirs;
+		const gchar * const *dirs;
 
 		for (dirs = gtk_source_language_manager_get_search_path (lm);
 		     dirs != NULL && *dirs != NULL;
@@ -240,19 +304,11 @@ _gtk_source_language_manager_get_rng_file (GtkSourceLanguageManager *lm)
 	return lm->priv->rng_file;
 }
 
-static void
-prepend_lang (G_GNUC_UNUSED gchar      *id,
-	      GtkSourceLanguage        *lang,
-	      GtkSourceLanguageManager *lm)
-{
-	lm->priv->available_languages =
-		g_slist_prepend (lm->priv->available_languages, g_object_ref (lang));
-}
-
 static void
 ensure_languages (GtkSourceLanguageManager *lm)
 {
 	GSList *filenames, *l;
+	GPtrArray *ids_array = NULL;
 
 	if (lm->priv->language_ids != NULL)
 		return;
@@ -260,8 +316,9 @@ ensure_languages (GtkSourceLanguageManager *lm)
 	lm->priv->language_ids = g_hash_table_new_full (g_str_hash, g_str_equal,
 							g_free, g_object_unref);
 
-	filenames = _gtk_source_view_get_file_list (gtk_source_language_manager_get_search_path (lm),
-						    LANG_FILE_SUFFIX);
+	filenames = _gtk_source_view_get_file_list ((gchar **)gtk_source_language_manager_get_search_path (lm),
+						    LANG_FILE_SUFFIX,
+						    TRUE);
 
 	for (l = filenames; l != NULL; l = l->next)
 	{
@@ -283,6 +340,11 @@ ensure_languages (GtkSourceLanguageManager *lm)
 			g_hash_table_insert (lm->priv->language_ids,
 					     g_strdup (lang->priv->id),
 					     lang);
+
+			if (ids_array == NULL)
+				ids_array = g_ptr_array_new ();
+
+			g_ptr_array_add (ids_array, g_strdup (lang->priv->id));
 		}
 		else
 		{
@@ -290,31 +352,40 @@ ensure_languages (GtkSourceLanguageManager *lm)
 		}
 	}
 
-	g_hash_table_foreach (lm->priv->language_ids, (GHFunc) prepend_lang, lm);
+	if (ids_array != NULL)
+	{
+		/* Ensure the array is NULL terminated */
+		g_ptr_array_add (ids_array, NULL);
+		lm->priv->ids = (gchar **)g_ptr_array_free (ids_array, FALSE);
+	}
 
 	g_slist_foreach (filenames, (GFunc) g_free, NULL);
 	g_slist_free (filenames);
 }
 
+
 /**
- * gtk_source_language_manager_get_available_languages:
- * @lm: a #GtkSourceLanguageManager.
+ * gtk_source_language_manager_get_language_ids:
+ * @lm: a #GtkSourceLanguageManager
  *
- * Gets a list of available languages for the given language manager.
+ * Returns the ids of the available languages.
  *
- * Returns: a list of #GtkSourceLanguage. Return value is owned by @lm and should
- * not be modified or freed.
- **/
-const GSList *
-gtk_source_language_manager_get_available_languages (GtkSourceLanguageManager *lm)
+ * Returns: a %NULL-terminated array of string containing the ids of the
+ * available languages or %NULL if no language is available. The array
+ * is owned by @lm and must not be modified.
+ */
+G_CONST_RETURN gchar* G_CONST_RETURN *
+gtk_source_language_manager_get_language_ids (GtkSourceLanguageManager *lm)
 {
 	g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE_MANAGER (lm), NULL);
+
 	ensure_languages (lm);
-	return lm->priv->available_languages;
+
+	return (const gchar * const *)lm->priv->ids;
 }
 
 /**
- * gtk_source_language_manager_get_language_by_id:
+ * gtk_source_language_manager_get_language:
  * @lm: a #GtkSourceLanguageManager.
  * @id: a language id.
  *
@@ -324,13 +395,16 @@ gtk_source_language_manager_get_available_languages (GtkSourceLanguageManager *l
  * Returns: a #GtkSourceLanguage, or %NULL if there is no language
  * identified by the given @id. Return value is owned by @lm and should not
  * be freed.
- **/
+ */
 GtkSourceLanguage *
-gtk_source_language_manager_get_language_by_id (GtkSourceLanguageManager *lm,
-						const gchar              *id)
+gtk_source_language_manager_get_language (GtkSourceLanguageManager *lm,
+					  const gchar              *id)
 {
+	g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE_MANAGER (lm), NULL);
 	g_return_val_if_fail (id != NULL, NULL);
+
 	ensure_languages (lm);
+
 	return g_hash_table_lookup (lm->priv->language_ids, id);
 }
 
