@@ -19,40 +19,44 @@
  */
 
 #include "gtksourceview-i18n.h"
-#include "gtksourcestylemanager.h"
+#include "gtksourcestyleschememanager.h"
 #include "gtksourceview.h"
 #include "gtksourcelanguage-private.h"
 #include "gtksourcestyle-private.h"
 #include <libxml/parser.h>
 #include <string.h>
 
-#define STYLE_TEXT		"text"
-#define STYLE_SELECTED		"text-selected"
-#define STYLE_BRACKET_MATCH	"bracket-match"
-#define STYLE_BRACKET_MISMATCH	"bracket-mismatch"
-#define STYLE_CURSOR		"cursor"
-#define STYLE_SECONDARY_CURSOR	"secondary-cursor"
-#define STYLE_CURRENT_LINE	"current-line"
-#define STYLE_LINE_NUMBERS	"line-numbers"
+#define STYLE_TEXT			"text"
+#define STYLE_SELECTED			"selection"
+#define STYLE_SELECTED_UNFOCUSED	"selection-unfocused"
+#define STYLE_BRACKET_MATCH		"bracket-match"
+#define STYLE_BRACKET_MISMATCH		"bracket-mismatch"
+#define STYLE_CURSOR			"cursor"
+#define STYLE_SECONDARY_CURSOR		"secondary-cursor"
+#define STYLE_CURRENT_LINE		"current-line"
+#define STYLE_LINE_NUMBERS		"line-numbers"
+#define STYLE_RIGHT_MARGIN		"right-margin"
 
-#define STYLE_SCHEME_VERSION	"1.0"
+#define STYLE_SCHEME_VERSION		"1.0"
 
-static const gchar *get_color_by_name  (GtkSourceStyleScheme *scheme,
-					const gchar          *name);
+#define DEFAULT_STYLE_SCHEME		"classic"
 
 
 enum {
 	PROP_0,
 	PROP_ID,
-	PROP_NAME
+	PROP_NAME,
+	PROP_DESCRIPTION,
+	PROP_FILENAME
 };
 
 struct _GtkSourceStyleSchemePrivate
 {
 	gchar *id;
 	gchar *name;
-	gchar *author;
+	GPtrArray *authors;
 	gchar *description;
+	gchar *filename;
 	GtkSourceStyleScheme *parent;
 	gchar *parent_id;
 	GHashTable *defined_styles;
@@ -70,7 +74,14 @@ gtk_source_style_scheme_finalize (GObject *object)
 	g_hash_table_destroy (scheme->priv->named_colors);
 	g_hash_table_destroy (scheme->priv->style_cache);
 	g_hash_table_destroy (scheme->priv->defined_styles);
-	g_free (scheme->priv->author);
+	g_free (scheme->priv->filename);
+
+	if (scheme->priv->authors != NULL)
+	{
+		g_ptr_array_foreach (scheme->priv->authors, (GFunc)g_free, NULL);
+		g_ptr_array_free (scheme->priv->authors, TRUE);
+	}
+
 	g_free (scheme->priv->description);
 	g_free (scheme->priv->id);
 	g_free (scheme->priv->name);
@@ -84,7 +95,7 @@ gtk_source_style_scheme_finalize (GObject *object)
 
 static void
 gtk_source_style_scheme_set_property (GObject 	   *object,
-				      guint 	    prop_id,
+				      guint         prop_id,
 				      const GValue *value,
 				      GParamSpec   *pspec)
 {
@@ -93,22 +104,15 @@ gtk_source_style_scheme_set_property (GObject 	   *object,
 
 	switch (prop_id)
 	{
-	    case PROP_ID:
-		tmp = scheme->priv->id;
-		scheme->priv->id = g_value_dup_string (value);
-		g_free (tmp);
-		break;
+		case PROP_ID:
+			tmp = scheme->priv->id;
+			scheme->priv->id = g_value_dup_string (value);
+			g_free (tmp);
+			break;
 
-	    case PROP_NAME:
-		tmp = scheme->priv->name;
-		scheme->priv->name = g_value_dup_string (value);
-		g_free (tmp);
-		g_object_notify (object, "name");
-		break;
-
-	    default:
-		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-		break;
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+			break;
 	}
 }
 
@@ -122,17 +126,25 @@ gtk_source_style_scheme_get_property (GObject 	 *object,
 
 	switch (prop_id)
 	{
-	    case PROP_ID:
-		    g_value_set_string (value, scheme->priv->id);
-		    break;
+		case PROP_ID:
+			g_value_set_string (value, scheme->priv->id);
+			break;
 
-	    case PROP_NAME:
-		    g_value_set_string (value, scheme->priv->name);
-		    break;
+		case PROP_NAME:
+			g_value_set_string (value, scheme->priv->name);
+			break;
 
-	    default:
-		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-		break;
+		case PROP_DESCRIPTION:
+			g_value_set_string (value, scheme->priv->description);
+			break;
+
+		case PROP_FILENAME:
+			g_value_set_string (value, scheme->priv->filename);
+			break;
+
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+			break;
 	}
 }
 
@@ -149,9 +161,7 @@ gtk_source_style_scheme_class_init (GtkSourceStyleSchemeClass *klass)
 	 * GtkSourceStyleScheme:id:
 	 *
 	 * Style scheme id, a unique string used to identify the style scheme
-	 * in #GtkSourceStyleManager.
-	 *
-	 * Since: 2.0
+	 * in #GtkSourceStyleSchemeManager.
 	 */
 	g_object_class_install_property (object_class,
 					 PROP_ID,
@@ -165,8 +175,6 @@ gtk_source_style_scheme_class_init (GtkSourceStyleSchemeClass *klass)
 	 * GtkSourceStyleScheme:name:
 	 *
 	 * Style scheme name, a translatable string to present to user.
-	 *
-	 * Since: 2.0
 	 */
 	g_object_class_install_property (object_class,
 					 PROP_NAME,
@@ -174,7 +182,33 @@ gtk_source_style_scheme_class_init (GtkSourceStyleSchemeClass *klass)
 						 	      _("Style scheme name"),
 							      _("Style scheme name"),
 							      NULL,
-							      G_PARAM_READWRITE));
+							      G_PARAM_READABLE));
+
+	/**
+	 * GtkSourceStyleScheme:name:
+	 *
+	 * Style scheme name, a translatable string to present to user.
+	 */
+	g_object_class_install_property (object_class,
+					 PROP_NAME,
+					 g_param_spec_string ("description",
+						 	      _("Style scheme description"),
+							      _("Style scheme description"),
+							      NULL,
+							      G_PARAM_READABLE));
+
+	/**
+	 * GtkSourceStyleScheme:filename:
+	 *
+	 * Style scheme filename or NULL.
+	 */
+	g_object_class_install_property (object_class,
+					 PROP_FILENAME,
+					 g_param_spec_string ("filename",
+						 	      _("Style scheme filename"),
+							      _("Style scheme filename"),
+							      NULL,
+							      G_PARAM_READABLE));
 
 	g_type_class_add_private (object_class, sizeof (GtkSourceStyleSchemePrivate));
 }
@@ -231,6 +265,58 @@ gtk_source_style_scheme_get_name (GtkSourceStyleScheme *scheme)
 	return scheme->priv->name;
 }
 
+/**
+ * gtk_source_style_scheme_get_description:
+ * @scheme: a #GtkSourceStyleScheme.
+ *
+ * Returns: @scheme description (if defined) or NULL.
+ *
+ * Since: 2.0
+ */
+const gchar *
+gtk_source_style_scheme_get_description (GtkSourceStyleScheme *scheme)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme), NULL);
+	return scheme->priv->description;
+}
+
+/**
+ * gtk_source_style_scheme_get_authors:
+ * @scheme: a #GtkSourceStyleScheme.
+ *
+ * Returns: a %NULL-terminated array containing the @scheme authors or
+ * %NULL if no author is specified by the style
+ * scheme.
+ *
+ * Since: 2.0
+ */
+G_CONST_RETURN gchar* G_CONST_RETURN *
+gtk_source_style_scheme_get_authors (GtkSourceStyleScheme *scheme)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme), NULL);
+
+	if (scheme->priv->authors == NULL)
+		return NULL;
+
+	return (G_CONST_RETURN gchar* G_CONST_RETURN *)scheme->priv->authors->pdata;
+}
+
+/**
+ * gtk_source_style_scheme_get_filename:
+ * @scheme: a #GtkSourceStyleScheme.
+ *
+ * Returns: @scheme file name if the scheme was created parsing a
+ * style scheme file or NULL in the other cases.
+ *
+ * Since: 2.0
+ */
+const gchar *
+gtk_source_style_scheme_get_filename (GtkSourceStyleScheme *scheme)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme), NULL);
+	return scheme->priv->filename;
+}
+
 /**
  * _gtk_source_style_scheme_new:
  * @id: scheme id.
@@ -255,91 +341,6 @@ _gtk_source_style_scheme_new (const gchar *id,
 	return scheme;
 }
 
-static GtkSourceStyle *
-fix_style_colors (GtkSourceStyleScheme *scheme,
-		  GtkSourceStyle       *real_style)
-{
-	GtkSourceStyle *style;
-
-	style = gtk_source_style_copy (real_style);
-
-	if (style->mask & GTK_SOURCE_STYLE_USE_BACKGROUND)
-	{
-		const gchar *color = get_color_by_name (scheme, style->background);
-
-		if (color == NULL)
-			/* warning is spit out in get_color_by_name,
-			 * here we make sure style doesn't have NULL color */
-			style->mask &= ~GTK_SOURCE_STYLE_USE_BACKGROUND;
-		else
-			style->background = g_intern_string (color);
-	}
-
-	if (style->mask & GTK_SOURCE_STYLE_USE_FOREGROUND)
-	{
-		const gchar *color = get_color_by_name (scheme, style->foreground);
-
-		if (color == NULL)
-			style->mask &= ~GTK_SOURCE_STYLE_USE_FOREGROUND;
-		else
-			style->foreground = g_intern_string (color);
-	}
-
-	return style;
-}
-
-/**
- * gtk_source_style_scheme_get_style:
- * @scheme: a #GtkSourceStyleScheme.
- * @style_name: style name to find.
- *
- * Returns: style which corresponds to @style_name in the @scheme,
- * or %NULL when no style with this name found. It is owned by @sheme
- * and may not be unref'ed.
- *
- * Since: 2.0
- */
-/*
- * It's a little weird because we have named colors: styles loaded from
- * scheme file can have "#red" or "blue", and we want to give out styles
- * which have nice colors suitable for gdk_color_parse(), so that GtkSourceStyle
- * foreground and background properties are the same as GtkTextTag's.
- * So, defined_styles hash has named colors; styles returned with get_style()
- * have real colors.
- */
-GtkSourceStyle *
-gtk_source_style_scheme_get_style (GtkSourceStyleScheme *scheme,
-				   const gchar          *style_name)
-{
-	GtkSourceStyle *style = NULL;
-	GtkSourceStyle *real_style;
-
-	g_return_val_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme), NULL);
-	g_return_val_if_fail (style_name != NULL, NULL);
-
-	if (g_hash_table_lookup_extended (scheme->priv->style_cache, style_name,
-					  NULL, (gpointer *) &style))
-		return style;
-
-	real_style = g_hash_table_lookup (scheme->priv->defined_styles, style_name);
-
-	if (real_style == NULL)
-	{
-		if (scheme->priv->parent != NULL)
-			style = gtk_source_style_scheme_get_style (scheme->priv->parent,
-								   style_name);
-		if (style != NULL)
-			g_object_ref (style);
-	}
-	else
-	{
-		style = fix_style_colors (scheme, real_style);
-	}
-
-	g_hash_table_insert (scheme->priv->style_cache, g_strdup (style_name), style);
-	return style;
-}
-
 /**
  * get_color_by_name:
  * @scheme: a #GtkSourceStyleScheme.
@@ -385,6 +386,100 @@ get_color_by_name (GtkSourceStyleScheme *scheme,
 	return color;
 }
 
+static GtkSourceStyle *
+fix_style_colors (GtkSourceStyleScheme *scheme,
+		  GtkSourceStyle       *real_style)
+{
+	GtkSourceStyle *style;
+	guint i;
+	struct {
+		guint mask;
+		guint offset;
+	} attributes[] = {
+		{ GTK_SOURCE_STYLE_USE_BACKGROUND, G_STRUCT_OFFSET (GtkSourceStyle, background) },
+		{ GTK_SOURCE_STYLE_USE_FOREGROUND, G_STRUCT_OFFSET (GtkSourceStyle, foreground) },
+		{ GTK_SOURCE_STYLE_USE_LINE_BACKGROUND, G_STRUCT_OFFSET (GtkSourceStyle, line_background) }
+	};
+
+	style = gtk_source_style_copy (real_style);
+
+	for (i = 0; i < G_N_ELEMENTS (attributes); i++)
+	{
+		if (style->mask & attributes[i].mask)
+		{
+			const gchar **attr = G_STRUCT_MEMBER_P (style, attributes[i].offset);
+			const gchar *color = get_color_by_name (scheme, *attr);
+
+			if (color == NULL)
+				/* warning is spit out in get_color_by_name,
+				 * here we make sure style doesn't have NULL color */
+				style->mask &= ~attributes[i].mask;
+			else
+				*attr = g_intern_string (color);
+		}
+	}
+
+	return style;
+}
+
+/**
+ * gtk_source_style_scheme_get_style:
+ * @scheme: a #GtkSourceStyleScheme.
+ * @style_id: id of the style to retrieve.
+ *
+ * Returns: style which corresponds to @style_id in the @scheme,
+ * or %NULL when no style with this name found. It is owned by @scheme
+ * and may not be unref'ed.
+ *
+ * Since: 2.0
+ */
+/*
+ * It's a little weird because we have named colors: styles loaded from
+ * scheme file can have "#red" or "blue", and we want to give out styles
+ * which have nice colors suitable for gdk_color_parse(), so that GtkSourceStyle
+ * foreground and background properties are the same as GtkTextTag's.
+ * Yet we do need to preserve what we got from file in style schemes,
+ * since there may be child schemes which may redefine colors or something,
+ * so we can't translate colors when loading scheme.
+ * So, defined_styles hash has named colors; styles returned with get_style()
+ * have real colors.
+ */
+GtkSourceStyle *
+gtk_source_style_scheme_get_style (GtkSourceStyleScheme *scheme,
+				   const gchar          *style_id)
+{
+	GtkSourceStyle *style = NULL;
+	GtkSourceStyle *real_style;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme), NULL);
+	g_return_val_if_fail (style_id != NULL, NULL);
+
+	if (g_hash_table_lookup_extended (scheme->priv->style_cache, style_id,
+					  NULL, (gpointer) &style))
+		return style;
+
+	real_style = g_hash_table_lookup (scheme->priv->defined_styles, style_id);
+
+	if (real_style == NULL)
+	{
+		if (scheme->priv->parent != NULL)
+			style = gtk_source_style_scheme_get_style (scheme->priv->parent,
+								   style_id);
+		if (style != NULL)
+			g_object_ref (style);
+	}
+	else
+	{
+		style = fix_style_colors (scheme, real_style);
+	}
+
+	g_hash_table_insert (scheme->priv->style_cache,
+			     g_strdup (style_id),
+			     style);
+
+	return style;
+}
+
 #if 0
 /**
  * gtk_source_style_scheme_set_style:
@@ -410,22 +505,20 @@ gtk_source_style_scheme_set_style (GtkSourceStyleScheme *scheme,
 }
 #endif
 
-/**
- * gtk_source_style_scheme_get_matching_brackets_style:
- * @scheme: a #GtkSourceStyleScheme.
- *
- * Returns: style which corresponds to "bracket-match" name, to use
- * in an editor. It is owned by @scheme and may not be unref'ed.
- *
- * Since: 2.0
- */
 GtkSourceStyle *
-gtk_source_style_scheme_get_matching_brackets_style (GtkSourceStyleScheme *scheme)
+_gtk_source_style_scheme_get_matching_brackets_style (GtkSourceStyleScheme *scheme)
 {
 	g_return_val_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme), NULL);
 	return gtk_source_style_scheme_get_style (scheme, STYLE_BRACKET_MATCH);
 }
 
+GtkSourceStyle *
+_gtk_source_style_scheme_get_right_margin_style (GtkSourceStyleScheme *scheme)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme), NULL);
+	return gtk_source_style_scheme_get_style (scheme, STYLE_RIGHT_MARGIN);
+}
+
 static gboolean
 get_color (GtkSourceStyle *style,
 	   gboolean        foreground,
@@ -463,19 +556,12 @@ get_color (GtkSourceStyle *style,
 	return FALSE;
 }
 
-/**
- * gtk_source_style_scheme_get_current_line_color:
- * @scheme: a #GtkSourceStyleScheme.
- * @color: a #GdkColor structure to fill.
- *
- * Returns: %TRUE if @scheme has style for current line set, or %FALSE
- * otherwise.
- *
- * Since: 2.0
+/*
+ * Returns TRUE if the style for current-line set in the scheme
  */
 gboolean
-gtk_source_style_scheme_get_current_line_color (GtkSourceStyleScheme *scheme,
-						GdkColor             *color)
+_gtk_source_style_scheme_get_current_line_color (GtkSourceStyleScheme *scheme,
+						 GdkColor             *color)
 {
 	GtkSourceStyle *style;
 
@@ -562,6 +648,10 @@ set_cursor_colors (GtkWidget      *widget,
 	if (strcmp (widget_name, gtk_widget_get_name (widget)) != 0)
 		gtk_widget_set_name (widget, widget_name);
 
+	g_object_set_data (G_OBJECT (widget),
+			   "gtk-source-view-cursor-color-set",
+			   GINT_TO_POINTER (TRUE));
+
 	g_free (rc_string);
 	g_free (widget_name);
 #else
@@ -573,9 +663,10 @@ static void
 unset_cursor_colors (GtkWidget *widget)
 {
 #if !GTK_CHECK_VERSION(2,11,3)
-	set_cursor_colors (widget,
-			   &widget->style->text[GTK_STATE_NORMAL],
-			   &widget->style->text_aa[GTK_STATE_NORMAL]);
+	if (g_object_get_data (G_OBJECT (widget), "gtk-source-view-cursor-color-set") != NULL)
+		set_cursor_colors (widget,
+				   &widget->style->text[GTK_STATE_NORMAL],
+				   &widget->style->text_aa[GTK_STATE_NORMAL]);
 #else
 	gtk_widget_modify_cursor (widget, NULL, NULL);
 #endif
@@ -634,13 +725,17 @@ _gtk_source_style_scheme_apply (GtkSourceStyleScheme *scheme,
 
 		style = gtk_source_style_scheme_get_style (scheme, STYLE_TEXT);
 		set_text_style (widget, style, GTK_STATE_NORMAL);
-		set_text_style (widget, style, GTK_STATE_ACTIVE);
 		set_text_style (widget, style, GTK_STATE_PRELIGHT);
 		set_text_style (widget, style, GTK_STATE_INSENSITIVE);
 
 		style = gtk_source_style_scheme_get_style (scheme, STYLE_SELECTED);
 		set_text_style (widget, style, GTK_STATE_SELECTED);
 
+		style2 = gtk_source_style_scheme_get_style (scheme, STYLE_SELECTED_UNFOCUSED);
+		if (style2 == NULL)
+			style2 = style;
+		set_text_style (widget, style2, GTK_STATE_ACTIVE);
+
 		style = gtk_source_style_scheme_get_style (scheme, STYLE_LINE_NUMBERS);
 		set_line_numbers_style (widget, style);
 
@@ -693,7 +788,7 @@ parse_style (GtkSourceStyleScheme *scheme,
 {
 	GtkSourceStyle *use_style = NULL;
 	GtkSourceStyle *result = NULL;
-	xmlChar *fg = NULL, *bg = NULL;
+	xmlChar *fg = NULL, *bg = NULL, *line_bg = NULL;
 	gchar *style_name = NULL;
 	guint mask = 0;
 	gboolean bold = FALSE;
@@ -741,6 +836,7 @@ parse_style (GtkSourceStyleScheme *scheme,
 
 	fg = xmlGetProp (node, BAD_CAST "foreground");
 	bg = xmlGetProp (node, BAD_CAST "background");
+	line_bg = xmlGetProp (node, BAD_CAST "line-background");
 	get_bool (node, "italic", &mask, GTK_SOURCE_STYLE_USE_ITALIC, &italic);
 	get_bool (node, "bold", &mask, GTK_SOURCE_STYLE_USE_BOLD, &bold);
 	get_bool (node, "underline", &mask, GTK_SOURCE_STYLE_USE_UNDERLINE, &underline);
@@ -748,7 +844,7 @@ parse_style (GtkSourceStyleScheme *scheme,
 
 	if (use_style)
 	{
-		if (fg != NULL || bg != NULL || mask != 0)
+		if (fg != NULL || bg != NULL || line_bg != NULL || mask != 0)
 		{
 			g_set_error (error, ERROR_QUARK, 0,
 				     "in style '%s': style attributes used along with use-style",
@@ -784,6 +880,12 @@ parse_style (GtkSourceStyleScheme *scheme,
 			result->background = g_intern_string ((char*) bg);
 			result->mask |= GTK_SOURCE_STYLE_USE_BACKGROUND;
 		}
+
+		if (line_bg != NULL)
+		{
+			result->line_background = g_intern_string ((char*) line_bg);
+			result->mask |= GTK_SOURCE_STYLE_USE_LINE_BACKGROUND;
+		}
 	}
 
 	*style_p = result;
@@ -851,7 +953,11 @@ parse_style_scheme_child (GtkSourceStyleScheme *scheme,
 	else if (strcmp ((char*) node->name, "author") == 0)
 	{
 		xmlChar *tmp = xmlNodeGetContent (node);
-		scheme->priv->author = g_strdup ((char*) tmp);
+		if (scheme->priv->authors == NULL)
+			scheme->priv->authors = g_ptr_array_new ();
+
+		g_ptr_array_add (scheme->priv->authors, g_strdup ((char*) tmp));
+
 		xmlFree (tmp);
 	}
 	else if (strcmp ((char*) node->name, "description") == 0)
@@ -935,6 +1041,10 @@ parse_style_scheme_element (GtkSourceStyleScheme *scheme,
 		if (node->type == XML_ELEMENT_NODE)
 			if (!parse_style_scheme_child (scheme, node, error))
 				return;
+
+	/* NULL-terminate the array of authors */
+	if (scheme->priv->authors != NULL)
+		g_ptr_array_add (scheme->priv->authors, NULL);
 }
 
 /**
@@ -992,6 +1102,8 @@ _gtk_source_style_scheme_new_from_file (const gchar *filename)
 	}
 
 	scheme = g_object_new (GTK_TYPE_SOURCE_STYLE_SCHEME, NULL);
+	scheme->priv->filename = g_strdup (filename);
+
 	parse_style_scheme_element (scheme, node, &error);
 
 	if (error != NULL)
@@ -1007,6 +1119,7 @@ _gtk_source_style_scheme_new_from_file (const gchar *filename)
 
 	xmlFreeDoc (doc);
 	g_free (text);
+
 	return scheme;
 }
 
@@ -1061,9 +1174,10 @@ _gtk_source_style_scheme_set_parent (GtkSourceStyleScheme *scheme,
 GtkSourceStyleScheme *
 _gtk_source_style_scheme_get_default (void)
 {
-	GtkSourceStyleManager *manager;
+	GtkSourceStyleSchemeManager *manager;
 
-	manager = gtk_source_style_manager_get_default ();
+	manager = gtk_source_style_scheme_manager_get_default ();
 
-	return gtk_source_style_manager_get_scheme (manager, "gvim");
+	return gtk_source_style_scheme_manager_get_scheme (manager,
+							   DEFAULT_STYLE_SCHEME);
 }
