@@ -34,6 +34,7 @@
 #include <gtksourceview/gtksourcelanguage.h>
 #include <gtksourceview/gtksourcelanguagemanager.h>
 #include <gtksourceview/gtksourcestyleschememanager.h>
+#include <gtksourceview/gtksourceprintcompositor.h>
 #ifdef TEST_XML_MEM
 #include <libxml/xmlreader.h>
 #endif
@@ -52,11 +53,16 @@ static GtkSourceStyleScheme *style_scheme = NULL;
 
 #define READ_BUFFER_SIZE   4096
 
+#define MARK_TYPE_1      "one"
+#define MARK_TYPE_2      "two"
+
 
 /* Private prototypes -------------------------------------------------------- */
 
 static void       open_file_cb                   (GtkAction       *action,
 						  gpointer         user_data);
+static void       print_file_cb                  (GtkAction       *action,
+						  gpointer         user_data);						  
 static void       debug_thing_3_cb		 (GtkAction       *action,
 						  gpointer         user_data);
 
@@ -64,6 +70,8 @@ static void       new_view_cb                    (GtkAction       *action,
 						  gpointer         user_data);
 static void       numbers_toggled_cb             (GtkAction       *action,
 						  gpointer         user_data);
+static void       marks_toggled_cb               (GtkAction       *action,
+						  gpointer         user_data);
 static void       margin_toggled_cb              (GtkAction       *action,
 						  gpointer         user_data);
 static void       hl_line_toggled_cb             (GtkAction       *action,
@@ -96,6 +104,8 @@ static GtkActionEntry buffer_action_entries[] = {
 
 static GtkActionEntry view_action_entries[] = {
 	{ "FileMenu", NULL, "_File" },
+	{ "Print", GTK_STOCK_PRINT, "_Print", "<control>P",
+	  "Print the current file", G_CALLBACK (print_file_cb) },
 	{ "ViewMenu", NULL, "_View" },
 	{ "NewView", GTK_STOCK_NEW, "_New View", NULL,
 	  "Create a new view of the file", G_CALLBACK (new_view_cb) },
@@ -110,6 +120,9 @@ static GtkToggleActionEntry toggle_entries[] = {
 	{ "ShowNumbers", NULL, "Show _Line Numbers", NULL,
 	  "Toggle visibility of line numbers in the left margin",
 	  G_CALLBACK (numbers_toggled_cb), FALSE },
+	{ "ShowMarks", NULL, "Show Line _Marks", NULL,
+	  "Toggle visibility of marks in the left margin",
+	  G_CALLBACK (marks_toggled_cb), FALSE },
 	{ "ShowMargin", NULL, "Show Right M_argin", NULL,
 	  "Toggle visibility of right margin indicator",
 	  G_CALLBACK (margin_toggled_cb), FALSE },
@@ -167,6 +180,7 @@ static const gchar *view_ui_description =
 "      <menuitem action=\"NewView\"/>"
 "      <separator/>"
 "      <menuitem action=\"ShowNumbers\"/>"
+"      <menuitem action=\"ShowMarks\"/>"
 "      <menuitem action=\"ShowMargin\"/>"
 "      <menuitem action=\"HlLine\"/>"
 "      <menuitem action=\"WrapLines\"/>"
@@ -205,6 +219,7 @@ static const gchar *buffer_ui_description =
 "  <menubar name=\"MainMenu\">"
 "    <menu action=\"FileMenu\">"
 "      <menuitem action=\"Open\"/>"
+"      <menuitem action=\"Print\"/>"
 "      <menuitem action=\"DebugThing3\"/>"
 "      <separator/>"
 "      <menuitem action=\"Quit\"/>"
@@ -280,6 +295,16 @@ gtk_source_buffer_load_file (GtkSourceBuffer *source_buffer,
 	return TRUE;
 }
 
+static void
+remove_all_marks (GtkSourceBuffer *buffer)
+{
+	GtkTextIter s, e;
+
+	gtk_text_buffer_get_bounds (GTK_TEXT_BUFFER (buffer), &s, &e);
+
+	gtk_source_buffer_remove_source_marks (buffer, &s, &e, NULL);
+}
+
 /* Note this is wrong for several reasons, e.g. g_pattern_match is broken
  * for glob matching. */
 static GtkSourceLanguage *
@@ -420,7 +445,7 @@ get_language (GtkTextBuffer *buffer, const gchar *filename)
 	GtkSourceLanguage *language = NULL;
 	GtkTextIter start, end;
 	gchar *text;
-	const gchar *lang_string;
+	gchar *lang_string;
 
 	gtk_text_buffer_get_start_iter (buffer, &start);
 	end = start;
@@ -466,6 +491,8 @@ open_file (GtkSourceBuffer *buffer, const gchar *filename)
 		g_free (curdir);
 	}
 
+	remove_all_marks (buffer);
+
 	success = gtk_source_buffer_load_file (buffer, filename, NULL);
 
 	if (!success)
@@ -529,6 +556,15 @@ numbers_toggled_cb (GtkAction *action, gpointer user_data)
 		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
 }
 
+static void
+marks_toggled_cb (GtkAction *action, gpointer user_data)
+{
+	g_return_if_fail (GTK_IS_TOGGLE_ACTION (action) && GTK_IS_SOURCE_VIEW (user_data));
+	gtk_source_view_set_show_line_marks (
+		GTK_SOURCE_VIEW (user_data),
+		gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)));
+}
+
 static void
 margin_toggled_cb (GtkAction *action, gpointer user_data)
 {
@@ -764,6 +800,216 @@ open_file_cb (GtkAction *action, gpointer user_data)
 	gtk_widget_destroy (chooser);
 }
 
+#define NON_BLOCKING_PAGINATION
+
+#ifndef NON_BLOCKING_PAGINATION
+
+static void
+begin_print (GtkPrintOperation        *operation, 
+	     GtkPrintContext          *context,
+	     GtkSourcePrintCompositor *compositor)
+{
+	gint n_pages;
+	
+	while (!gtk_source_print_compositor_paginate (compositor, context))
+		;
+
+	n_pages = gtk_source_print_compositor_get_n_pages (compositor);
+	gtk_print_operation_set_n_pages (operation, n_pages);
+}
+
+#else
+
+static gboolean
+paginate (GtkPrintOperation        *operation, 
+	  GtkPrintContext          *context,
+	  GtkSourcePrintCompositor *compositor)
+{
+	g_print ("Pagination progress: %.2f %%\n", gtk_source_print_compositor_get_pagination_progress (compositor) * 100.0);
+	
+	if (gtk_source_print_compositor_paginate (compositor, context))
+	{
+		gint n_pages;
+
+		g_assert (gtk_source_print_compositor_get_pagination_progress (compositor) == 1.0);
+		g_print ("Pagination progress: %.2f %%\n", gtk_source_print_compositor_get_pagination_progress (compositor) * 100.0);
+		        
+		n_pages = gtk_source_print_compositor_get_n_pages (compositor);
+		gtk_print_operation_set_n_pages (operation, n_pages);
+
+
+		
+		return TRUE;
+	}
+     
+	return FALSE;
+}
+
+#endif
+
+#define ENABLE_CUSTOM_OVERLAY
+
+static void
+draw_page (GtkPrintOperation        *operation,
+	   GtkPrintContext          *context,
+	   gint                      page_nr,
+	   GtkSourcePrintCompositor *compositor)
+{
+#ifdef ENABLE_CUSTOM_OVERLAY
+
+	/* This part of the code shows how to add a custom overlay to the
+	   printed text generated by GtkSourcePrintCompositor */
+
+	cairo_t *cr;
+	PangoLayout *layout;
+	PangoFontDescription *desc;
+	PangoRectangle rect;
+
+	
+	cr = gtk_print_context_get_cairo_context (context);
+	
+	cairo_save (cr);
+	
+	layout = gtk_print_context_create_pango_layout (context);
+
+	pango_layout_set_text (layout, "Draft", -1);
+
+	desc = pango_font_description_from_string ("Sans Bold 120");
+	pango_layout_set_font_description (layout, desc);
+	pango_font_description_free (desc);
+
+	
+	pango_layout_get_extents (layout, NULL, &rect);
+		
+  	cairo_move_to (cr, 
+  		       (gtk_print_context_get_width (context) - ((double) rect.width / (double) PANGO_SCALE)) / 2,
+  		       (gtk_print_context_get_height (context) - ((double) rect.height / (double) PANGO_SCALE)) / 2);
+	  		       
+	pango_cairo_layout_path (cr, layout);
+	
+  	/* Font Outline */
+	cairo_set_source_rgba (cr, 0.85, 0.85, 0.85, 0.80);
+	cairo_set_line_width (cr, 0.5);
+	cairo_stroke_preserve (cr);
+
+	/* Font Fill */
+	cairo_set_source_rgba (cr, 0.8, 0.8, 0.8, 0.60);
+	cairo_fill (cr);
+	  	  
+	g_object_unref (layout);
+	cairo_restore (cr);
+#endif
+
+	/* To print page_nr you only need to call the following function */
+	gtk_source_print_compositor_draw_page (compositor, context, page_nr);
+}
+
+static void
+end_print (GtkPrintOperation        *operation, 
+	   GtkPrintContext          *context,
+	   GtkSourcePrintCompositor *compositor)
+{
+	g_object_unref (compositor);
+}
+
+#define LINE_NUMBERS_FONT_NAME	"Sans 8"
+#define HEADER_FONT_NAME	"Sans 11"
+#define FOOTER_FONT_NAME	"Sans 11"
+#define BODY_FONT_NAME		"Monospace 9"
+	
+/*
+#define SETUP_FROM_VIEW 
+*/
+
+#undef SETUP_FROM_VIEW 
+
+
+static void
+print_file_cb (GtkAction *action, gpointer user_data)
+{
+	GtkSourceView *view;
+	GtkSourceBuffer *buffer;
+	GtkSourcePrintCompositor *compositor;
+	GtkPrintOperation *operation;	
+	const gchar *filename;
+	gchar *basename;
+	
+	g_return_if_fail (GTK_IS_SOURCE_VIEW (user_data));
+	
+	view = GTK_SOURCE_VIEW (user_data);
+
+	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view))); 
+	
+	filename = g_object_get_data (G_OBJECT (buffer), "filename");	
+	basename = g_filename_display_basename (filename);
+	
+#ifdef SETUP_FROM_VIEW
+	compositor = gtk_source_print_compositor_new_from_view (view);
+#else
+	
+	compositor = gtk_source_print_compositor_new (buffer);
+
+	gtk_source_print_compositor_set_tab_width (compositor,
+						   gtk_source_view_get_tab_width (view));
+
+	gtk_source_print_compositor_set_wrap_mode (compositor,
+						   gtk_text_view_get_wrap_mode (GTK_TEXT_VIEW (view)));
+
+	gtk_source_print_compositor_set_print_line_numbers (compositor, 1);
+
+	gtk_source_print_compositor_set_body_font_name (compositor,
+							BODY_FONT_NAME);
+
+	/* To test line numbers font != text font */
+	gtk_source_print_compositor_set_line_numbers_font_name (compositor,
+								LINE_NUMBERS_FONT_NAME);
+
+	gtk_source_print_compositor_set_header_format (compositor,
+						       TRUE,
+						       "Printed on %A",
+						       "test-widget",
+						       "%F");
+
+	gtk_source_print_compositor_set_footer_format (compositor,
+						       TRUE,
+						       "%T",
+						       basename,
+						       "Page %N/%Q");
+
+	gtk_source_print_compositor_set_print_header (compositor, TRUE);
+	gtk_source_print_compositor_set_print_footer (compositor, TRUE);
+
+	gtk_source_print_compositor_set_header_font_name (compositor,
+							  HEADER_FONT_NAME);
+
+	gtk_source_print_compositor_set_footer_font_name (compositor,
+							  FOOTER_FONT_NAME);
+#endif	
+	operation = gtk_print_operation_new ();
+			
+	gtk_print_operation_set_job_name (operation, basename);
+	
+	gtk_print_operation_set_show_progress (operation, TRUE);
+
+#ifndef NON_BLOCKING_PAGINATION	
+  	g_signal_connect (G_OBJECT (operation), "begin-print", 
+			  G_CALLBACK (begin_print), compositor);
+#else
+  	g_signal_connect (G_OBJECT (operation), "paginate", 
+			  G_CALLBACK (paginate), compositor);
+#endif
+	g_signal_connect (G_OBJECT (operation), "draw-page", 
+			  G_CALLBACK (draw_page), compositor);
+	g_signal_connect (G_OBJECT (operation), "end-print", 
+			  G_CALLBACK (end_print), compositor);
+
+	gtk_print_operation_run (operation, 
+				 GTK_PRINT_OPERATION_ACTION_PRINT_DIALOG,
+				 NULL, NULL);
+
+	g_object_unref (operation);
+	g_free (basename);
+}
 
 /* View UI callbacks ------------------------------------------------------------------ */
 
@@ -857,6 +1103,72 @@ window_deleted_cb (GtkWidget *widget, GdkEvent *ev, gpointer user_data)
 	return TRUE;
 }
 
+static gboolean
+button_press_cb (GtkWidget *widget, GdkEventButton *ev, gpointer user_data)
+{
+	GtkSourceView *view;
+	GtkSourceBuffer *buffer;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_VIEW (widget), FALSE);
+
+	view = GTK_SOURCE_VIEW (widget);
+	buffer = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view)));
+
+	if (!gtk_source_view_get_show_line_marks (view))
+		return FALSE;
+
+	/* check that the click was on the left gutter */
+	if (ev->window == gtk_text_view_get_window (GTK_TEXT_VIEW (view),
+						    GTK_TEXT_WINDOW_LEFT))
+	{
+		gint y_buf;
+		GtkTextIter line_start;
+		GSList *mark_list;
+		const gchar *mark_type;
+
+		if (ev->button == 1)
+			mark_type = MARK_TYPE_1;
+		else
+			mark_type = MARK_TYPE_2;
+
+		gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (view),
+						       GTK_TEXT_WINDOW_LEFT,
+						       ev->x, ev->y,
+						       NULL, &y_buf);
+
+		/* get line bounds */
+		gtk_text_view_get_line_at_y (GTK_TEXT_VIEW (view),
+					     &line_start,
+					     y_buf,
+					     NULL);
+
+		/* get the marks already in the line */
+		mark_list = gtk_source_buffer_get_source_marks_at_line (buffer,
+									gtk_text_iter_get_line (&line_start),
+									mark_type);
+
+		if (mark_list != NULL)
+		{
+			/* just take the first and delete it */
+			gtk_text_buffer_delete_mark (GTK_TEXT_BUFFER (buffer),
+						     GTK_TEXT_MARK (mark_list->data));
+		}
+		else
+		{
+			/* no mark found: create one */
+			gtk_source_buffer_create_source_mark (buffer,
+							      NULL,
+							      mark_type,
+							      &line_start);
+		}
+
+		g_slist_free (mark_list);
+	}
+
+	return FALSE;
+}
+
+
 /* Window creation functions -------------------------------------------------------- */
 
 static GtkWidget *
@@ -864,6 +1176,7 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 {
 	GtkWidget *window, *sw, *view, *vbox, *pos_label, *menu;
 	PangoFontDescription *font_desc = NULL;
+	GdkPixbuf *pixbuf;
 	GtkAccelGroup *accel_group;
 	GtkActionGroup *action_group;
 	GtkUIManager *ui_manager;
@@ -884,8 +1197,9 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 	if (style_scheme)
 		gtk_source_buffer_set_style_scheme (buffer, style_scheme);
 
-	g_signal_connect (buffer, "mark_set", G_CALLBACK (move_cursor_cb), view);
+	g_signal_connect (buffer, "mark-set", G_CALLBACK (move_cursor_cb), view);
 	g_signal_connect (buffer, "changed", G_CALLBACK (update_cursor_position), view);
+	g_signal_connect (view, "button-press-event", G_CALLBACK (button_press_cb), NULL);
 	g_signal_connect (window, "delete-event", (GCallback) window_deleted_cb, view);
 
 	/* action group and UI manager */
@@ -958,6 +1272,10 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_show_line_numbers (from));
 
+		action = gtk_action_group_get_action (action_group, "ShowMarks");
+		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
+					      gtk_source_view_get_show_line_marks (from));
+
 		action = gtk_action_group_get_action (action_group, "ShowMargin");
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 					      gtk_source_view_get_show_right_margin (from));
@@ -993,6 +1311,35 @@ create_view_window (GtkSourceBuffer *buffer, GtkSourceView *from)
 		g_free (tmp);
 	}
 
+	/* add source mark pixbufs */
+	error = NULL;
+	if ((pixbuf = gdk_pixbuf_new_from_file (DATADIR "/pixmaps/apple-green.png", &error)))
+	{
+		gtk_source_view_set_mark_category_pixbuf (GTK_SOURCE_VIEW (view), MARK_TYPE_1, pixbuf);
+		gtk_source_view_set_mark_category_priority (GTK_SOURCE_VIEW (view), MARK_TYPE_1, 1);
+		g_object_unref (pixbuf);
+	}
+	else
+	{
+		g_message ("could not load source mark 1 image. Spurious messages might get triggered: %s\n",
+		error->message);
+		g_error_free (error);
+	}
+
+	error = NULL;
+	if ((pixbuf = gdk_pixbuf_new_from_file (DATADIR "/pixmaps/apple-red.png", &error)))
+	{
+		gtk_source_view_set_mark_category_pixbuf (GTK_SOURCE_VIEW (view), MARK_TYPE_2, pixbuf);
+		gtk_source_view_set_mark_category_priority (GTK_SOURCE_VIEW (view), MARK_TYPE_2, 2);
+		g_object_unref (pixbuf);
+	}
+	else
+	{
+		g_message ("could not load source mark 2 image. Spurious messages might get triggered: %s\n",
+		error->message);
+		g_error_free (error);
+	}
+
 	gtk_widget_show_all (vbox);
 
 	return window;
@@ -1035,6 +1382,9 @@ create_main_window (GtkSourceBuffer *buffer)
 	action = gtk_action_group_get_action (action_group, "ShowNumbers");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
 
+	action = gtk_action_group_get_action (action_group, "ShowMarks");
+	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), TRUE);
+
 	action = gtk_action_group_get_action (action_group, "ShowMargin");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), FALSE);
 
